import calendar
import threading
from datetime import datetime
from email.utils import formatdate
from http.cookies import SimpleCookie
from inspect import getsource
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

import base62  # type: ignore
import requests
from attr import attrib, dataclass

from .type_alias import TypeType

_local_session = threading.local()


def requests_session() -> requests.Session:
    """Wsgidav uses threading, lets get a session per thread."""
    local_dict = _local_session.__dict__
    session = local_dict.get("session")
    if not session:
        local_dict["session"] = session = requests.Session()
    return session


def cattrib(
    attrib_type: Optional[TypeType] = None,
    check: Optional[Callable] = None,
    optional: bool = False,
    **kwargs,
):
    if "default" in kwargs and kwargs["default"] is None:
        optional = True

    def handler(object, attribute, value):
        if value is None and optional:
            return value
        # Some python versions cannot do the check assume True
        is_inst = True
        try:
            if attrib_type is not None:
                is_inst = isinstance(value, attrib_type)
        except TypeError:
            pass
        if not is_inst:
            raise TypeError(
                f"{attribute.name} ({type(value)}) is not of type {attrib_type}"
            )
        if check and not check(value):
            source = getsource(check).strip()
            raise ValueError(f"check failed: {source}")
        return value

    return attrib(validator=handler, on_setattr=handler, **kwargs)


def get_rfc1123_time(secs: Optional[float] = None) -> str:
    """Return <secs> in rfc 1123 date/time format (pass secs=None for current date)."""
    return formatdate(timeval=secs, localtime=False, usegmt=True)


def to_string(data: bytes) -> str:
    return base62.encodebytes(data)


def from_string(data: str) -> bytes:
    return base62.decodebytes(data)


@dataclass
class AppInfo:
    start_response: Callable = cattrib(check=lambda x: callable(x))
    environ: Dict[str, Any] = cattrib(dict)
    secure: bool = cattrib(bool, default=True)


def set_cookie(
    info: AppInfo,
    key: str,
    value: str,
    ttl: int,
    status: int,
    headers: List[Tuple[str, str]],
    exc_info=None,
):
    cookie: SimpleCookie = SimpleCookie()
    cookie[key] = value
    date = datetime.utcnow()
    unixtime = calendar.timegm(date.utctimetuple())
    cookie[key]["expires"] = get_rfc1123_time(unixtime + ttl)
    if info.secure:
        cookie[key]["secure"] = True
        cookie[key]["httponly"] = True
    headers.append(cast(Tuple[str, str], tuple(str(cookie).split(": "))))
    info.start_response(status, headers, exc_info)



import fcntl
import json
import shelve
import threading
import traceback
import weakref
from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from pathlib import Path
from typing import Callable

import psycopg2.extensions
from psycopg2 import InterfaceError, OperationalError, connect
from psycopg2.extensions import connection as PsycopgConnection
from wsgidav.lock_man.lock_storage import LockStorageDict
from wsgidav.util import get_module_logger

_logger = get_module_logger(__name__)


class ManabiTimeoutMixin:
    _max_timeout: float = 0

    @property
    def max_timeout(self) -> float:
        return self._max_timeout

    @max_timeout.setter
    def max_timeout(self, value: float) -> None:
        if value <= 0:
            raise ValueError("max_timeout must be a positive integer.")
        self._max_timeout = value

    def set_timeout(self, lock):
        max_timeout = self.max_timeout
        timeout = lock.get("timeout")

        if not timeout:
            lock["timeout"] = max_timeout
        else:
            if timeout > max_timeout:
                lock["timeout"] = max_timeout


class ManabiContextLockMixin(ABC):
    @abstractmethod
    def acquire(self):
        pass

    @abstractmethod
    def release(self):
        pass

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()


class ManabiShelfLock(ManabiContextLockMixin):
    _storage_object: Callable[[], "ManabiShelfLockLockStorage"]

    def __init__(self, storage_path, storage_object: "ManabiShelfLockLockStorage"):
        self._storage_path = storage_path
        # type manually checked
        self._storage_object = weakref.ref(storage_object)  # type: ignore
        self._semaphore = 0
        self._lock_file = open(f"{storage_path}.lock", "wb+")
        self._fd = self._lock_file.fileno()
        self.acquire_write = self.acquire_read = self.acquire
        self._id = -1

    def acquire(self):
        tid = threading.get_ident()
        if self._id and self._id != tid:
            _logger.error("Do not use from multiple threads")
        self._id = tid
        if self._semaphore == 0:
            fcntl.flock(self._fd, fcntl.LOCK_EX)
            self._storage_object()._dict = shelve.open(str(self._storage_path))
        self._semaphore += 1

    def release(self):
        tid = threading.get_ident()
        if self._semaphore == 0:
            _logger.error(
                f"Inconsistent use of lock. {''.join(traceback.format_stack())}"
            )
        if self._id and self._id != tid:
            _logger.error("Do not use from multiple threads")
        self._id = tid
        self._semaphore -= 1
        if self._semaphore == 0:
            storage_object = self._storage_object()
            if storage_object._dict is not None:
                storage_object._dict.close()
                storage_object._dict = None
            fcntl.flock(self._fd, fcntl.LOCK_UN)
        self._id = threading.get_ident()


class ManabiShelfLockLockStorage(LockStorageDict, ManabiTimeoutMixin):
    def __init__(self, refresh: float, storage: Path):
        super().__init__()
        self.max_timeout = refresh / 2
        self._storage = storage
        # this is a cast
        self._lock: ManabiShelfLock = ManabiShelfLock(storage, self)  # type:ignore

    def open(self):
        pass

    def close(self):
        pass

    def create(self, path, lock):
        with self._lock:
            self.set_timeout(lock)
            super().create(path, lock)

    def clear(self):
        if self._dict:
            with self._lock:
                self._dict.clear()

    def refresh(self, token, *, timeout):
        with self._lock:
            return super().refresh(token, timeout=timeout)

    def get(self, token):
        with self._lock:
            return super().get(token)

    def delete(self, token):
        with self._lock:
            return super().delete(token)

    def get_lock_list(self, path, *, include_root, include_children, token_only):
        with self._lock:
            return super().get_lock_list(
                path,
                include_root=include_root,
                include_children=include_children,
                token_only=token_only,
            )


class ManabiPostgresLock(ManabiContextLockMixin):
    _storage_object: Callable[[], "ManabiDbLockStorage"]

    def __init__(self, storage_object: "ManabiDbLockStorage"):
        self._id = None
        # type manually checked
        self._storage_object = weakref.ref(storage_object)  # type: ignore
        self.acquire_write = self.acquire_read = self.acquire
        self._semaphore = 0
        self._id = -1

    def acquire(self):
        tid = threading.get_ident()
        if self._id and self._id != tid:
            _logger.error("Do not use from multiple threads")
        self._id = tid
        if self._semaphore == 0:
            _logger.info(f"{tid} acquire")
            self._storage_object().execute(
                "LOCK TABLE manabi_lock IN ACCESS EXCLUSIVE MODE"
            )
        self._semaphore += 1

    def release(self):
        tid = threading.get_ident()
        if self._semaphore == 0:
            _logger.error(
                f"Inconsistent use of lock. {''.join(traceback.format_stack())}"
            )
        if self._id and self._id != tid:
            _logger.error("Do not use from multiple threads")
        self._id = tid
        self._semaphore -= 1
        if self._semaphore == 0:
            _logger.info(f"{tid} release")
            self._storage_object()._connection.commit()


class ManabiPostgresDict(MutableMapping):
    _storage_object: Callable[[], "ManabiDbLockStorage"]

    def __init__(self, storage_object: "ManabiDbLockStorage", lock):
        # type manually checked
        self._storage_object = weakref.ref(storage_object)  # type: ignore
        self._lock = lock

    @staticmethod
    def decode_lock(lock):
        if "owner" in lock:
            owner = lock["owner"]
            if isinstance(owner, bytes):
                lock["owner"] = {"manabi_was_bytes": owner.decode("UTF-8")}

    @staticmethod
    def encode_lock(lock):
        if "owner" in lock:
            owner = lock["owner"]
            if "manabi_was_bytes" in owner:
                lock["owner"] = owner["manabi_was_bytes"].encode("UTF-8")

    def cleanup(self):
        with self._lock:
            self._storage_object().execute("DELETE FROM manabi_lock;")

    def __len__(self):
        with self._lock:
            cursor = self._storage_object().execute("SELECT count(*) FROM manabi_lock")
            return int(cursor.fetchone()[0])

    def __iter__(self):
        with self._lock:
            cursor = self._storage_object().execute("SELECT token FROM manabi_lock")
            for token in cursor.fetchall():
                yield token[0]

    def __delitem__(self, token):
        with self._lock:
            self._storage_object().execute(
                "DELETE FROM manabi_lock WHERE token = %s", (str(token),)
            )

    def __contains__(self, token):
        with self._lock:
            cursor = self._storage_object().execute(
                "SELECT 1 FROM manabi_lock WHERE token = %s", (str(token),)
            )
            if cursor.fetchone() is None:
                return False
            else:
                return True

    def __setitem__(self, token, lock):
        with self._lock:
            self.decode_lock(lock)
            json_lock = json.dumps(lock)
            self._storage_object().execute(
                """
                    INSERT INTO manabi_lock(token, data) VALUES (%(token)s, %(data)s)
                        ON CONFLICT(token) DO
                        UPDATE SET data = %(data)s WHERE manabi_lock.token = %(token)s
                """,
                {
                    "token": token,
                    "data": json_lock,
                },
            )

    def __getitem__(self, token):
        with self._lock:
            cursor = self._storage_object().execute(
                "SELECT data FROM manabi_lock WHERE token = %s", (str(token),)
            )
            lock = cursor.fetchone()
            if lock is None:
                raise KeyError(f"{token} not found")
            lock = lock[0]
            self.encode_lock(lock)
            return lock


class ManabiDbLockStorage(LockStorageDict, ManabiTimeoutMixin):
    _connection: PsycopgConnection

    def __init__(self, refresh: float, postgres_dsn: str):
        super().__init__()
        self._postgres_dsn = postgres_dsn
        self.max_timeout = refresh / 2
        self.connect()
        # this is a cast
        self._lock: ManabiPostgresLock = ManabiPostgresLock(self)  # type: ignore
        self._dict = ManabiPostgresDict(self, self._lock)

    def connect(self):
        try:
            if self._connection:
                # it the connecton failed, this might cause an exception, we do not care.
                self._connection.close()
        except Exception:
            pass
        self._connection = connect(self._postgres_dsn)
        self._connection.commit()
        self._connection.autocommit = False
        self._connection.set_session(
            isolation_level=psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE
        )
        self._cursor = self._connection.cursor()

    def open(self):
        pass

    def execute(self, *args, **kwargs):
        try:
            self._cursor.execute(*args, **kwargs)
        except (InterfaceError, OperationalError):
            _logger.error(
                f"Postgres connection lost, reconnecting. {''.join(traceback.format_stack())}"
            )

            self.connect()
            self._cursor.execute(*args, **kwargs)
        return self._cursor

    def close(self):
        pass

    def create(self, path, lock):
        with self._lock:
            self.set_timeout(lock)
            super().create(path, lock)



from functools import partial
from http.cookies import SimpleCookie
from typing import Any, Callable, Dict, List
from unittest.mock import MagicMock

from wsgidav.mw.base_mw import BaseMiddleware

from .token import Config, State, Token
from .util import AppInfo, get_rfc1123_time, set_cookie

_error_message_403 = """
<html>
    <head><title>403 Forbidden</title></head>
    <body>
        <h1>403 Forbidden</h1>
        {0}
    </body>
</html>
""".strip()


class ManabiAuthenticator(BaseMiddleware):
    # Instead of accepting an override for HTTPAuthenticator as for everything else,
    # wsgidav just expects a class extending HTTPAuthenticator in the middleware_stack.
    # In it also accesses the domain_controller with out check if it exists in some debug code.
    # https://github.com/mar10/wsgidav/pull/204
    def get_domain_controller(self) -> Any:
        return MagicMock()

    def manabi_secure(self) -> bool:
        config = self.config
        if "manabi" not in config:
            return True
        manabi = config["manabi"]
        if "secure" not in manabi:
            return True
        return manabi["secure"]

    def access_denied(self, start_response: Callable, reason: str = "") -> List[bytes]:
        body = _error_message_403
        content = body.format(reason).encode("UTF-8")
        start_response(
            "403 Forbidden",
            [
                ("Content-Type", "text/html"),
                ("Content-Length", str(len(content))),
                ("Date", get_rfc1123_time()),
            ],
        )
        return [content]

    def update_env(self, info: AppInfo, token: Token, id_: str):
        environ = info.environ
        # Update the path for security, so we can't ever be tricked into serving a
        # path not authenticated by the token.
        path = token.path_as_url()
        environ["PATH_INFO"] = path
        environ["REQUEST_URI"] = path
        environ["manabi.path"] = path

        environ["wsgidav.auth.user_name"] = f"{path.strip('/')}|{id_[10:18]}"
        environ["manabi.token"] = token

    def refresh(self, id_: str, info: AppInfo, token: Token, ttl: int):
        new = Token.from_token(token)
        self.update_env(info, token, id_)
        return self.next_app(
            info.environ,
            partial(set_cookie, info, id_, new.encode(), ttl),
        )

    def __call__(
        self, environ: Dict[str, Any], start_response: Callable
    ) -> List[bytes]:
        """__call__ is the entry-point of a wsgi-middleware.

        The method checks if the token is valid, so wsgi-dav can serve the document.
        It also refreshes the token by setting a new token in a cookie.

        Middlwares and wsgi-handlers are identical. Middlewares need to wrap
        start_response with a closure, if they want to add headers.
        https://www.python.org/dev/peps/pep-3333/
        """
        info = AppInfo(start_response, environ, self.manabi_secure())
        config = Config.from_dictionary(environ["wsgidav.config"])
        path_info = environ["PATH_INFO"]
        id_, _, suffix = path_info.strip("/").partition("/")
        suffix = suffix.strip("/")
        # We need this because we replace PATH_INFO with token.path, so later
        # we don't know if this was a directory access
        environ["manabi.dir_access"] = suffix == ""
        if not id_:
            return self.access_denied(start_response, "no token supplied")
        initial = Token.from_ciphertext(config.key, id_)
        check = initial.check()

        if check == State.invalid:
            return self.access_denied(start_response, check.value[1])

        cookie = environ.get("HTTP_COOKIE")
        ttl = config.ttl.refresh
        if cookie:
            cookie = SimpleCookie(cookie)
            refresh = cookie.get(initial.ciphertext)
            if refresh and refresh.value:
                refresh = Token.from_ciphertext(config.key, refresh.value)
                if refresh.refresh(config.ttl) == State.valid:
                    return self.refresh(id_, info, refresh, ttl)

        check = initial.initial(config.ttl)
        if initial.initial(config.ttl) == State.valid:
            return self.refresh(id_, info, initial, ttl)
        return self.access_denied(start_response, check.value[1])



from pathlib import Path

import wsgidav


def fix():
    # Workaround for https://github.com/python/mypy/issues/8545
    try:
        (Path(wsgidav.__file__).absolute().parent / "py.typed").touch()
    except Exception:
        pass


if __name__ == "__main__":
    fix()



# Generated by Django 3.2.15 on 2022-08-05 15:13

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Lock',
            fields=[
                ('token', models.CharField(max_length=128, primary_key=True, serialize=False)),
                ('data', models.JSONField()),
            ],
            options={
                'db_table': 'manabi_lock',
            },
        ),
    ]



import calendar
import struct
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Optional, Tuple, Union

import umsgpack  # type: ignore
from attr import Factory, dataclass
from branca import Branca  # type: ignore

from .type_alias import OptionalProp, PropType
from .util import cattrib, from_string


def now() -> int:
    return calendar.timegm(datetime.utcnow().timetuple())


@dataclass
class TTL:
    initial: int = cattrib(int)
    refresh: int = cattrib(int)

    @classmethod
    def from_dictionary(cls, config: dict) -> "TTL":
        initial = config["manabi"]["initial"]
        refresh = config["manabi"]["refresh"]
        return cls(initial, refresh)


@dataclass
class Key:
    data: bytes = cattrib(bytes, lambda x: len(x) == 32)

    @classmethod
    def from_dictionary(cls, config: dict) -> "Key":
        return cls(from_string(config["manabi"]["key"]))


@dataclass
class Config:
    key: Key = cattrib(Key)
    ttl: TTL = cattrib(TTL)

    @classmethod
    def from_dictionary(cls, config: dict) -> "Config":
        return cls(Key.from_dictionary(config), TTL.from_dictionary(config))


class State(Enum):
    valid = 1, "the token is intact, path is valid and ttl ok"
    expired = 2, "the token is intact, path valid but the ttl is expired"
    not_valid_yet = 3, "the token intact, but not valid yet, clockskew?"
    invalid = 4, "the token is not valid, authentication failed"


@dataclass
class Token:
    key: Key = cattrib(Key)
    path: Optional[Path] = cattrib(Path, default=None)
    payload: OptionalProp = cattrib(OptionalProp, default=None)
    timestamp: Optional[int] = cattrib(
        int, default=Factory(lambda: now()), optional=True
    )
    ciphertext: str = cattrib(str, default=None)

    def as_url(self) -> str:
        if self.path is None:
            raise ValueError("path may not be None")
        path = self.path.name
        return f"{self.encode()}/{path}"

    def path_as_url(self) -> str:
        return f"/{self.path}"

    def encode(self) -> str:
        if self.path is None:
            raise ValueError("path may not be None")
        if self.timestamp is None:
            self.timestamp = now()
        self.ciphertext = _encode(
            self.key.data,
            str(self.path),
            self.payload,
            self.timestamp,
        )
        return self.ciphertext

    @classmethod
    def from_token(cls, token: "Token", timestamp: Optional[int] = None) -> "Token":
        if timestamp is None:
            return cls(token.key, token.path, token.payload, now())
        else:
            return cls(token.key, token.path, token.payload, timestamp)

    @classmethod
    def from_ciphertext(cls, key: Key, ciphertext: str) -> "Token":
        assert ciphertext
        branca = Branca(key.data)
        try:
            timestamp = branca.timestamp(ciphertext)
        except (struct.error, ValueError):
            return cls(key, None, None)
        try:
            token_path, token_payload = _decode(branca, ciphertext)
        except RuntimeError:
            return cls(key, None, timestamp)
        return cls(key, token_path, token_payload, timestamp, ciphertext)

    def check(self, ttl: Optional[int] = None) -> State:
        if self.path is None or self.timestamp is None:
            return State.invalid
        if ttl is not None:
            future = self.timestamp + ttl
            past = self.timestamp - 1
            t = now()
            if t < past:
                return State.not_valid_yet
            if t > future:
                return State.expired
        return State.valid

    def refresh(self, ttl: TTL) -> State:
        return self.check(ttl.refresh)

    def initial(self, ttl: TTL) -> State:
        return self.check(ttl.initial)


def _encode(
    key: bytes,
    path: str,
    payload: OptionalProp = None,
    now: Optional[int] = None,
) -> str:
    f = Branca(key)
    p = umsgpack.packb((path.encode("UTF-8"), payload))
    ciphertext = f.encode(p, now)
    return ciphertext


def _decode(
    key: Union[bytes, Branca],
    ciphertext: str,
    ttl=None,
) -> Tuple[Path, PropType]:
    if isinstance(key, Branca):
        f = key
    else:
        f = Branca(key)
    tpb, token_payload = umsgpack.unpackb(f.decode(ciphertext, ttl))
    token_path = tpb.decode("UTF-8")
    return Path(token_path), token_payload
